/*
 * This module implements Rio's main data structure, the registry.
 *
 * The registry keeps track of all pages in memory that need to be synced
 * to disk during a crash.
 *
 * A page gets an entry in the registry when it gets written in the buffer
 * cache (bwrite/bdwrite/bawrite) OR when it's mapped writable in a user
 * process's address space (vm_fault or pmap_protect).  Note that a page
 * may be in the object cache but not be in the registry (e.g. when the
 * page is read-only in a process's address space).
 *
 * The registry entry for a page is deleted when the page leaves the object
 * cache.  This happens at vm_page_remove.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_kern.h>
#include <vm/vm_extern.h>
#include <vm/vm_prot.h>
#include <vm/vm_object.h>
#include <vm/vm_page.h>
#include <vm/lock.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <sys/buf.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <machine/md_var.h>

#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#include <ufs/ffs/fs.h>
#include <miscfs/specfs/specdev.h>

#include <sys/disklabel.h>

#include <sys/rio.h>

#include <sys/conf.h>

#include <ufs/ffs/ffs_extern.h>

struct rio_reg *rio_reg;
enum {UNMALLOCED, MALLOCED, INITIALIZED} rio_reg_status = UNMALLOCED;

/*
 * prototypes for internal functions
 */
static void rio_update_buf __P((char *, struct buf *, vm_prot_t, int,
										int));
static int buf_is_ufs(char *, struct buf *bp);
static int vp_is_ufs(char *, struct vnode *);
static int check_range(char *, int);
static int compare_reg_entry(const void *, const void *);

/*
 * prototypes for external functions (that aren't normally used by rest of
 * kernel)
 */
struct disk;

vm_offset_t vnode_pager_addr __P((struct vnode *, vm_ooffset_t, int *));
int wd_setup __P((dev_t, struct disk **, struct disklabel **));
int wd_write __P((char *, long, daddr_t, dev_t, struct disk *,
						struct disklabel *));
int sd_setup __P((dev_t, struct disklabel **));
int sd_write __P((char *, long, daddr_t, dev_t, struct disklabel *));

extern char btext;

/*
 * Initialize Rio registry
 */
void
rio_init()
{
	int i;

	printf("initializing Rio registry with %d entries, each %d bytes\n",
		Maxmem, sizeof(struct rio_reg));
	rio_reg = (struct rio_reg *)
		    kmem_alloc(kernel_map, Maxmem*sizeof(struct rio_reg));
	rio_reg_status = MALLOCED;
	
	printf("rio_reg at address 0x%x\n", (long) rio_reg);

	for (i=0; i<Maxmem; i++) {
		rio_remove(i<<PAGE_SHIFT);
	}
	rio_reg_status = INITIALIZED;

	printf("&btext=0x%x, &etext=0x%x\n", &btext, &etext);
}

/*
 * Mark buffer as changing.
 * This is done before an explicit modification to the page by the kernel,
 *   and so is used only for pages in the buffer cache.
 * Unprotect pages to allow kernel to write to it.
 */
void
rio_before_mod(struct buf *bp)
{
	int i;
	int page_num;

	if (rio_reg_status != INITIALIZED) {
		panic("rio_before_mod while rio_reg is unintialized");
	}

	if (!buf_is_ufs("rio_before_mod", bp)) {
		return;
	}

	/* 
	 * sanity check to make sure rio_before_mod isn't called twice
	 * in a row (without an intervening rio_after_mod)
	 */
	for (i = 0; i < bp->b_npages; i++) {
		if (bp->b_pages[i] != NULL) {
			page_num = VM_PAGE_TO_PHYS(bp->b_pages[i])>>PAGE_SHIFT;
			if (rio_reg[page_num].flags & RIO_REG_CHANGING) {
				printf("warning: rio_before_mod: page 0x%x marked as changing, flags=0x%x\n",
					page_num, rio_reg[page_num].flags);
			}
		} else {
			printf("warning: rio_before_mod: b_pages[%d]=NULL\n",
				i);
		}
	}

	rio_update_buf("rio_before_mod", bp, VM_PROT_ALL, RIO_REG_CHANGING, ~0);
}

/*
 * Mark buffer as done being changed.
 * This is done after an explicit modification to the page by the kernel,
 *   and so is used only for pages in the buffer cache.
 * Reprotect page.
 */
void
rio_after_mod(struct buf *bp)
{
	int i;
	int page_num;

	if (rio_reg_status != INITIALIZED) {
		panic("rio_after_mod while Rio registry is unintialized");
	}

	if (!buf_is_ufs("rio_after_mod", bp)) {
		return;
	}

	/* 
	 * sanity check to make sure rio_before_mod was called before
	 * rio_after_mod
	 */
	for (i = 0; i < bp->b_npages; i++) {
		if (bp->b_pages[i] != NULL) {
			page_num = VM_PAGE_TO_PHYS(bp->b_pages[i])>>PAGE_SHIFT;
			if (! (rio_reg[page_num].flags & RIO_REG_CHANGING) ) {
				printf("error: rio_after_mod: page 0x%x not marked as changing, flags=0x%x\n",
					page_num, rio_reg[page_num].flags);
				panic("page not marked as changing");
			}
		} else {
			printf("warning: rio_after_mod: b_pages[%d]=NULL\n",
				i);
		}
	}

	rio_update_buf("rio_after_mod", bp, VM_PROT_READ|VM_PROT_EXECUTE,
			RIO_REG_VALID|RIO_REG_DIRTY, ~RIO_REG_CHANGING);
}

/*
 * Remove entry in Rio registry.  Called when a page is disassociated from an
 * object (currently only by vm_page_remove and when initializing Rio
 * registry).
 */
void
rio_remove(vm_offset_t phys_addr)
{
	int page_num = phys_addr >> PAGE_SHIFT;	/* physical page number */

	/* can't remove entries before the space is allocated */
	if (rio_reg_status == UNMALLOCED) {
		return;
	}

	if (check_range("rio_remove", page_num)) {
		return;
	}

	/*
	if (rio_reg_status==INITIALIZED &&
	    (rio_reg[page_num].flags & RIO_REG_VALID) ) {
		printf("info: rio_remove: removing rio_reg entry 0x%x\n",
			page_num);
	}
	*/

	rio_reg[page_num].dev = NODEV;
	rio_reg[page_num].blkno = -1;

	/*
	 * flags should be !RIO_REG_VALID, !RIO_REG_DIRTY, !RIO_REG_CHANGING,
	 * !RIO_REG_MMAP
	 */
	rio_reg[page_num].flags = 0;
	rio_reg[page_num].size = 0;
}

/*
 * Add mmap entry to Rio registry.  This only needs to be called for
 * pages that are mapped writable.
 */
void
rio_mmap(vm_page_t m)
{
	int page_num;
	struct vnode *vp;
	int do_next_flag = 0;

	/* can't add entries before the space is allocated */
	if (rio_reg_status == UNMALLOCED) {
		/*
		printf("warning: rio_mmap called before rio_reg was initialized\n");
		*/
		return;
	}

	/* sanity checks */
	if (m == NULL) {
		printf("error: rio_mmap: m is NULL\n");
		return;
	} else if (m->object == NULL) {
		printf("error: rio_mmap: m->object is NULL\n");
		return;
	} else if (m->object->type != OBJT_VNODE) {
		/*
		printf("info: rio_mmap: object type %d\n", m->object->type);
		*/
		return;
	}

	vp = m->object->handle;

	if (!vp_is_ufs("rio_mmap", vp)) {
		return;
	}

	page_num = VM_PAGE_TO_PHYS(m) >> PAGE_SHIFT;
	if (check_range("rio_mmap", page_num)) {
		return;
	}

	/*
	 * Don't bother adding to the registry if it's already there (to save
	 * time).  Pages are always removed before being assigned to a new
	 * object/offset.
	 */
	if (rio_reg[page_num].flags & RIO_REG_VALID) {
		return;
	}

	/*
	printf("info: rio_mmap: adding page 0x%x\n", page_num);
	vprint("rio_mmap", vp);
	*/

	rio_reg[page_num].dev = VTOI(vp)->i_dev;

	rio_reg[page_num].blkno =
		vnode_pager_addr(vp, IDX_TO_OFF(m->pindex), NULL);

	/*
	 * Check if block is not yet allocated in file system (blkno -1).
	 * This can happen if you mmap a hole in the file.  Allocate these
	 * now (rather than wait 'til later, e.g. at msync) so they're
	 * persistent right away.
	 */
	if (rio_reg[page_num].blkno == -1) {
		/*
		 * Use VOP_WRITE to allocate the file block for this page.
		 * This is how the normal kernel allocates the file block
		 * (when the kernel calls sync).
		 * I think this will also fill in the rest of the file block
		 * with the right data (zeroes in this case, since we're in a
		 * file hole).
		 */
		struct uio tmp_uio;
		struct iovec tmp_iov;
		struct ucred cred;
		int error;

		/*
		 * ffs_write wants a real credential (not NOCRED or FSCRED).
		 */
		cred.cr_uid = 0;

		tmp_uio.uio_iov = &tmp_iov;
		tmp_uio.uio_iovcnt = 1;
		tmp_uio.uio_offset = m->pindex * PAGE_SIZE;
		tmp_uio.uio_resid = PAGE_SIZE;
		tmp_uio.uio_segflg = UIO_NOCOPY;
		tmp_uio.uio_rw = UIO_WRITE;
		tmp_uio.uio_procp = NULL;

		/*
		 * iov_base doesn't matter, since we're using UIO_NOCOPY
		 */
		tmp_iov.iov_base = NULL;
		tmp_iov.iov_len = PAGE_SIZE;

		error = VOP_WRITE(vp, &tmp_uio, 0, &cred);
		if (error) {
			printf("rio_mmap's VOP_WRITE returned %d while filling in hole\n", error);
			panic("rio_mmap");
		}
	}
	rio_reg[page_num].size = PAGE_SIZE;
	rio_reg[page_num].flags |=
			(RIO_REG_VALID | RIO_REG_DIRTY | RIO_REG_MMAP);

}

/*
 * Dump the Rio registry to disk.  Called by panic.  Should exit with cpu_reset
 * (don't return, otherwise panic() will try to sync and wipe out the disk).
 */
void
rio_dump()
{
	int i;
	dev_t dev, last_dev;
	struct disklabel *label_ptr;
	struct disk *du;
	struct disklabel *lp;
	int numpages=0;
	struct vm_page tmp_page;
	vm_page_t tmp_page_ptr = &tmp_page;

	printf("dumping Rio registry to disk\n");

	/*
	 * Sort Rio registry.
	 */
	printf("sorting %d entries\n", Maxmem);
	for (i=0; i<Maxmem; i++) {
		rio_reg[i].pagenum = i;
	}
	qsort(rio_reg, Maxmem, sizeof(struct rio_reg), compare_reg_entry);
	printf("done sorting\n");

	/*
	 * Disable interrupts before doing dump
	 */
	splhigh();

	last_dev = NODEV;
	for (i=0; i<Maxmem; i++) {
		if ( (rio_reg[i].flags & RIO_REG_VALID) &&
			    (rio_reg[i].flags & RIO_REG_DIRTY) &&
			    rio_reg[i].blkno == -1) {
			printf("error: blkno -1\n");
		}

		if ( (rio_reg[i].flags & RIO_REG_VALID) &&
			    (rio_reg[i].flags & RIO_REG_DIRTY) &&
			    rio_reg[i].blkno != -1) {

			/* map the registry page */
			tmp_page.phys_addr = rio_reg[i].pagenum * PAGE_SIZE;
			pmap_qenter( (vm_offset_t) CADDR1, &tmp_page_ptr, 1);

			/*
			 * If first I/O to a device, set up device.
			 */
			if (rio_reg[i].dev != last_dev) {

				printf("setting up device 0x%x\n",
					rio_reg[i].dev);

				/*
				 * Got device major number from the file
				 * device.i386.  There should be a constant
				 * defined somewhere that defines these numbers.
				 */
				if (major(rio_reg[i].dev) == 0) {
					printf("\nIDE device\n");
					if (wd_setup(rio_reg[i].dev, &du, &lp)){
						/* maybe we should try harder,
						    e.g. by trying to dump
						    to the other devices */
						printf("error in setting up device\n");
					}
				} else if (major(rio_reg[i].dev) == 4) {
					printf("\nSCSI device\n");
					if (sd_setup(rio_reg[i].dev, &lp)) {
						/* maybe we should try harder,
						    e.g. by trying to dump
						    to the other devices */
						printf("error in setting up device\n");
					}
				} else {
					printf("error: undefined device 0x%x, major# %d, page %d\n",
						rio_reg[i].dev,
						major(rio_reg[i].dev),
						rio_reg[i].pagenum);
				}
				printf("done setting up device 0x%x\n",
					rio_reg[i].dev);
			}

			if (major(rio_reg[i].dev) == 0) {
				if (wd_write(CADDR1, (int) rio_reg[i].size,
						rio_reg[i].blkno,
						rio_reg[i].dev, du, lp)) {
					printf("error in writing to block %d\n",
						rio_reg[i].blkno);
				}
			} else if (major(rio_reg[i].dev) == 4) {
				if (sd_write(CADDR1, (int) rio_reg[i].size,
						rio_reg[i].blkno,
						rio_reg[i].dev, lp)) {
					printf("error in writing to block %d\n", rio_reg[i].blkno);
				}
			}

			numpages++;
			last_dev = rio_reg[i].dev;
		}
	}

	printf("dumped %d pages\n", numpages);
	cpu_reset();
}

/* Internal routines */

/*
 * Update status of pages for a buffer
 */
static void
rio_update_buf(char *message, struct buf *bp, vm_prot_t protect,
		    int orFlags, int andFlags)
{
	int i;
	int page_num;
	int status;
	char buf[PAGE_SIZE];

	if (bp->b_flags & (B_INVAL | B_ERROR) ) {
		printf("info: %s: B_INVAL or B_ERROR, b_flags=0x%x\n",
		    message, bp->b_flags);
		return;
	}

	if (round_page(bp->b_bcount) != (bp->b_npages << PAGE_SHIFT) ) {
		/*
		 * One set of routines that causes b_npages not to match
		 * b_bcount is getpbuf.  These can be safely ignored if
		 * bp->b_npages==0.  And I think these shouldn't occur since
		 * I'm not instrumenting reads via biodone.
		 */
		printf("error: %s: b_npages (%d) doesn't match b_count (%d)\n",
			message, bp->b_npages, bp->b_bcount);
		printf("b_flags=0x%x, b_error=%d, b_bufsize=%d, b_resid=%d, b_dev=0x%x, b_lblkno=0x%x, b_blkno=0x%x\n",
			bp->b_flags, bp->b_error, bp->b_bufsize, bp->b_resid,
			bp->b_dev, bp->b_lblkno, bp->b_blkno);
		for (i = 0; i*PAGE_SIZE < bp->b_bcount; i++) {
			printf("bp->b_pages[%d] = 0x%x\n", i, bp->b_pages[i]);
		}
			
		panic("b_npages doesn't match b_count");
	}

	for (i = 0; i < bp->b_npages; i++) {
		if (bp->b_pages[i] == NULL) {
			printf("error: %s: bp->b_pages[%d] is NULL\n",
			    message, i);
			panic("b_pages is NULL");
		}

		page_num = VM_PAGE_TO_PHYS(bp->b_pages[i]) >> PAGE_SHIFT;

		if (!check_range(message, page_num)) {
			/*
			if ( (orFlags & RIO_REG_VALID) &&
			    (! (rio_reg[page_num].flags & RIO_REG_VALID) ) ) {
				printf("info: %s: new rio_reg entry 0x%x\n",
					message, page_num);
			}
			*/

			/*
			 * Set information for entries that are valid.
			 * Even if the entry was valid before, some of the info
			 * might have changed (e.g. blkno, when a fragment
			 * grows).
			 */
			if (orFlags & RIO_REG_VALID) {

				/*
				 * Set blkno
				 */
				if (bp->b_lblkno == bp->b_blkno) {
					status = VOP_BMAP(bp->b_vp,
							bp->b_lblkno, NULL,
							&bp->b_blkno, NULL,
							NULL);
					if (status) {
						printf("warning: %s: VOP_BMAP returned %d\n",
							message, status);
					}
				}

				rio_reg[page_num].blkno =
					bp->b_blkno + i*PAGE_SIZE/DEV_BSIZE;
				if (rio_reg[page_num].blkno <= 0) {
					printf("warning: %s: rio_reg[0x%x].blkno = 0x%x\n",
						message, page_num,
						rio_reg[page_num].blkno);
					vprint(message, bp->b_vp);
				}

				/*
				 * Set device
				 */
				if (bp->b_dev == NODEV) {
					rio_reg[page_num].dev =
						VTOI(bp->b_vp)->i_dev;
				} else {
					rio_reg[page_num].dev = bp->b_dev;
				}

				/*
				 * The last page may not be full.
				 */
				if (i == bp->b_npages - 1) {
					rio_reg[page_num].size = bp->b_bcount -
					    PAGE_SIZE * (bp->b_npages - 1);
				} else {
					rio_reg[page_num].size = PAGE_SIZE;
				}
			}

			/*
			 * change protection of *virtual* page to protect.
			 * virtual address is based on bp->b_data
			 */
			status = vm_map_protect(buffer_map,
				    (vm_offset_t) bp->b_data + i*PAGE_SIZE,
				    (vm_offset_t) bp->b_data + (i+1)*PAGE_SIZE,
				    protect, FALSE);
			if (status != KERN_SUCCESS) {
				printf("error: %s, vm_map_protect returned %d\n",
					message, status);
			}

			/*
			 * Set valid flag last so we don't have half-baked data
			 * in a valid buffer.
			 */
			rio_reg[page_num].flags &= andFlags;
			rio_reg[page_num].flags |= orFlags;
		}
	}
}

/*
 * Make sure buffer belongs to a UFS file system.
 */
static int
buf_is_ufs(char *message, struct buf *bp)
{
	if (bp == NULL) {
		printf("info: %s: bp is NULL\n", message);
		return(0);
	} else {
		return(vp_is_ufs(message, bp->b_vp));
	}
}

/*
 * Make sure vnode belongs to a UFS file system.
 */
static int
vp_is_ufs(char *message, struct vnode *vp)
{
	if (vp == NULL) {
		/* printf("info: %s: vp is NULL\n", message); */
		return(0);
	} else if (vp->v_mount == NULL) {
		/* printf("info: %s: vp->v_mount is NULL\n", message); */
		return(0);
	} else if (vp->v_mount->mnt_stat.f_type != MOUNT_UFS) {
		/* printf("info: %s: buffer not UFS (f_type=%d)\n",
		    message, vp->v_mount->mnt_stat.f_type); */
		return(0);
	}
	return(1);
}

/*
 * Make sure page is in the range of physical memory.
 */
static int
check_range(char *message, int page_num)
{
	if (page_num >= Maxmem) {
		/*
		 * exiting X unmaps a bunch of strange memory addresses,
		 * so don't print an error message for rio_remove
		 */
		if (strcmp(message, "rio_remove")) {
			printf("warning: check_range (%s): page_num=0x%x\n",
				message, page_num);
		}
		/* panic("page_num out of range"); */
		return(1);
	} else {
		return(0);
	}
}

/*
 * Used by rio_dump to sort Rio registry entries.  Sorts by device,
 * then by blkno.
 */
static int
compare_reg_entry(const void *s1_ptr, const void *s2_ptr)
{
	struct rio_reg *entry_ptr1 = (struct rio_reg *) s1_ptr;
	struct rio_reg *entry_ptr2 = (struct rio_reg *) s2_ptr;

	if (entry_ptr1->dev < entry_ptr2->dev) {
		return -1;
	} else if (entry_ptr1->dev > entry_ptr2->dev) {
		return 1;
	} else if (entry_ptr1->blkno < entry_ptr2->blkno) {
		return -1;
	} else if (entry_ptr1->blkno > entry_ptr2->blkno) {
		return 1;
	} else {
		return 0;
	}
}
